/**
 * Copyright (C) 2008 Ovea <dev@testatoo.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.testatoo.container.impl;

import com.sun.grizzly.arp.AsyncHandler;
import com.sun.grizzly.arp.DefaultAsyncHandler;
import com.sun.grizzly.comet.CometAsyncFilter;
import com.sun.grizzly.http.SelectorThread;
import com.sun.grizzly.http.embed.GrizzlyWebServer;
import com.sun.grizzly.http.servlet.deployer.WebAppAdapter;
import com.sun.grizzly.http.webxml.WebappLoader;
import com.sun.grizzly.http.webxml.schema.WebApp;
import com.sun.grizzly.tcp.http11.GrizzlyAdapter;
import com.sun.grizzly.tcp.http11.GrizzlyRequest;
import com.sun.grizzly.tcp.http11.GrizzlyResponse;
import com.sun.grizzly.websockets.WebSocketAsyncFilter;
import org.testatoo.container.ContainerConfiguration;
import org.testatoo.container.ContainerSkeleton;

import java.io.File;
import java.util.Map;
import java.util.Set;

import static org.testatoo.container.util.ClassloaderUtils.tryEnhanceContextClassloaderWithClasspath;

public final class GrizzlyContainer extends ContainerSkeleton<GrizzlyWebServer> {

    volatile boolean running;

    public GrizzlyContainer(ContainerConfiguration properties) {
        super(properties);
    }

    protected void start(GrizzlyWebServer server) {
        try {
            server.start();
            running = true;
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    protected void stop(GrizzlyWebServer server) {
        try {
            server.stop();
            running = false;
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    protected boolean isRunning(GrizzlyWebServer server) {
        return running;
    }

    protected GrizzlyWebServer buildServer() {
        try {
            if (settings().hasServerClassPath()) {
                tryEnhanceContextClassloaderWithClasspath(settings().serverClassPath());
            }

            String root = webappRoot().getCanonicalPath();

            // create server
            GrizzlyWebServer ws = new GrizzlyWebServer(port(), root);
            SelectorThread st = ws.getSelectorThread();

            // Comet and WebSocket support
            AsyncHandler asyncHandler = new DefaultAsyncHandler();
            asyncHandler.addAsyncFilter(new CometAsyncFilter());
            asyncHandler.addAsyncFilter(new WebSocketAsyncFilter());
            st.setAsyncHandler(asyncHandler);
            st.setEnableAsyncExecution(true);

            // create default webapp
            WebApp defaultWebapp = new WebApp();

            // create webapp
            final WebApp webApp = WebappLoader.load(new File(webappRoot(), "WEB-INF/web.xml").getCanonicalPath());

            // add welcome-file handler
            if (!webApp.getWelcomeFileList().isEmpty() && !webApp.getWelcomeFileList().get(0).getWelcomeFile().isEmpty()) {
                ws.addGrizzlyAdapter(new GrizzlyAdapter() {
                    {
                        setHandleStaticResources(true);
                    }

                    @Override
                    public void service(GrizzlyRequest request, GrizzlyResponse response) throws Exception {
                        if ("/".equals(request.getRequestURI())) {
                            response.sendRedirect(webApp.getWelcomeFileList().get(0).getWelcomeFile().get(0));
                        } else {
                            super.service(request.getRequestURI(), request.getRequest(), response.getResponse());
                        }
                    }
                }, new String[]{"/"});
            }

            WebAppAdapter adapter = new WebAppAdapter(root, contextPath(), webApp, Thread.currentThread().getContextClassLoader(), defaultWebapp);
            for (Map.Entry<GrizzlyAdapter, Set<String>> entry : adapter.getToRegister().entrySet()) {
                ws.addGrizzlyAdapter(entry.getKey(), entry.getValue().toArray(new String[entry.getValue().size()]));
            }

            return ws;
        } catch (RuntimeException e) {
            throw e;
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }
}
